/**
 * Genetic algorithm solution to Andys football team problem.
 *
 * Steve
 * ---
 * www.steve.org.uk
 *
 */

#include <algorithm>
#include <string>

#include <stdio.h>
#include <stdlib.h>


#include "players.h"
#include "solution.h"



/**
 * Compare two objects by fitness - this is used to sort our
 * array of solutions.
 */
bool cmp_fitness( CSolution a, CSolution b)
{
  return ( a.getFitness() < b.getFitness() );
}



/**
 * Entry point to the code.
 *
 */
int main( int argc, char *argv[] )
{
  /* Seed our random number generator */
  srand( time(NULL) );

  /* The array of solutions / organisms we work with. */
  CSolution m_solutions[ 500 ];

  /* The number of generations to work for. */
  int nGenerations = 100;

  /* 
   *  If there's an argument assume it's the number of generations
   * to run for.
   */
  if ( argc == 2 )
  {
    nGenerations = atoi( argv[1] );
  }

  
  /*
   * Seed initial set.
   */
  for( int i = 0; i < 500; i++ )
  {
    m_solutions[i].seedRandomly();
  }


  /**
   * Run the evolution loop for the given number of generations.
   */
  for( int g = 0; g <= nGenerations; g++ )
  {

    /*
     * Sort the solutions so that we may discard the least fit.
     */
    std::sort(m_solutions, m_solutions+500, cmp_fitness );

    /* Display the current fittest solution. */
    printf( "Generation: %3d - Best : %6d\n", 
	      g, m_solutions[ 499 ].getFitness() );


    /*
     * Nuke the first 200 solutions - just replace them
     * with new random selections.
     */
    for( int i = 0; i < 200; i++ )
    {
      m_solutions[ i ] = CSolution();
      m_solutions[ i ].seedRandomly();
    }

    /*
     * Copy the best 100 solutions positions 200-400 
     */
    std::copy( m_solutions+400, m_solutions+500, m_solutions+200);
    std::copy( m_solutions+400, m_solutions+500, m_solutions+300);

    /*
     * Now randomly breed the solutions.
     */
    for( int i = 0; i < 500; i ++ )
    {
      int partner = rand()%500;
	
      m_solutions[ i ] = m_solutions[i].breed( m_solutions[ partner ]  );
    }
  }


  /*
   * Sort the solutions so we can display the fitest - our winner.
   */
  std::sort(m_solutions, m_solutions+500, cmp_fitness );

  printf( "\n" );
  if ( m_solutions[ 499 ].isValid() )
  {
    m_solutions[ 499 ].display();
  }
  else
  {
    printf("Best solution was invalid :(\n" );
  }

  return 0;
}
