/**
 * This object holds a collection of 11 players which make up
 * a solution.
 *
 * Note that the solution may not be valid ..
 *
 */


#include <string>
#include <map>

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "solution.h"
#include "players.h"

/* Player definitions. */
#include "playdef.cc"



/**
 * Constructor
 */
CSolution::CSolution()
{
}


/**
 * Destructor.
 */
CSolution::~CSolution()
{
}


/**
 * Is the solution we've chosen valid?
 * If the cost of all the players is over 50 it's not.
 * If more than two players are on the same team it's not.
 * Otherwise it is.
 */
int CSolution::isValid()
{
    int price = 0;

    /* Total each players cost. */
    for( int i = 0; i < 11; i++ )
    {
        price += m_players[i].getPrice();
    }

    /* Blown the budget - not valid */
    if ( price > 50 )
    {
        return(0);
    }

    /* Build up team counts. */
    std::map<std::string, int> teams;
    for( int i = 0; i < 11; i++ )
    {
        std::string team = m_players[i].getClub();
	++teams[team];
    }

    std::map<std::string, int>::iterator it;
    for (it = teams.begin(); it != teams.end(); ++it)
    {
        if ( it->second > 2 )
        {
            /* Found a team with more than two players. */
	    return 0;
	}
    }


    /* Make sure each player only occurs once. */
    std::map<std::string, int> players;
    for( int i = 0; i < 11; i++ )
    {
        std::string player = m_players[i].getName();
	++players[player];
    }

    for (it = players.begin(); it != players.end(); ++it)
    {
        if ( it->second > 1 )
        {
	  /* Found a team with duplicate players. */
	  return 0;
	}
    }


    {

      /*
       * Now check that the team has the relevent number of
       * players in each position.
       *
       * This code is suboptimal, although obviously correct..
       */
      int nGoalkeeprs = 0;
      int nForwards = 0;
      int nMidField = 0;
      int nCentreBacks = 0;
      int nFullBacks   = 0;

      for( int i = 0; i < 11; i++ )
      {
	  std::string position = m_players[i].getPosition();
	  if ( strcmp( position.c_str(), "Goal Keeper" ) == 0 )
	      nGoalkeeprs += 1;
	  if ( strcmp( position.c_str(), "Full Backs" ) == 0 )
	      nFullBacks += 1;
	  if ( strcmp( position.c_str(), "Centre Backs" ) == 0 )
	      nCentreBacks += 1;
	  if ( strcmp( position.c_str(), "Mid Field" ) == 0 )
	      nMidField += 1;
	  if ( strcmp( position.c_str(), "Forwards" ) == 0 )
	      nForwards += 1;
      }

      if ( nGoalkeeprs != 1 )
          return 0;

      if ( nFullBacks != 2 )
	  return 0;

      if ( nCentreBacks != 2 )
	  return 0;

      if ( nMidField != 4 )
	  return 0;

      if ( nForwards != 2 )
	  return 0;
    }

    /*
     * All our tests passed - so this *is* a valid team.
     *
     * W00t!
     */
    return 1;
}


/**
 * Calculate the fitness of this solution.
 * In our case the fitness is the thing that we're trying to increase,
 * so it's just the total of the points each player has.
 */
int CSolution::getFitness()
{
    int fitness = 0;
    int price   = 0;

    if ( ! isValid() )
    {
        return 0;
    }


    /* Total each players points + cost. */
    for( int i = 0; i < 11; i++ )
    {
        CPlayer player = m_players[i];

	fitness += player.getPoints();
	price   += player.getPrice();
    }

    return( fitness );
}


/**
 * Build up a team by randomly choosing 11 players.
 */
void CSolution::seedRandomly()
{
  /* Pick goalie. */
  int nGoalies = (sizeof(g_Goal_Keepers ) / sizeof(g_Goal_Keepers[0] ) );
  m_players[ 0 ] = g_Goal_Keepers[ rand() % nGoalies ];

  int nFullbacks = (sizeof(g_Full_Backs ) / sizeof(g_Full_Backs[0] ) );
  m_players[ 1 ] = g_Full_Backs[ rand() % nFullbacks ];
  m_players[ 2 ] = g_Full_Backs[ rand() % nFullbacks ];

  int nCenters = (sizeof(g_Centre_Backs ) / sizeof(g_Centre_Backs[0] ) );
  m_players[ 3 ] = g_Centre_Backs[ rand() % nCenters ];
  m_players[ 4 ] = g_Centre_Backs[ rand() % nCenters ];

  int nMiddle  = (sizeof(g_Mid_Field ) / sizeof(g_Mid_Field[0] ) );
  m_players[ 5 ] = g_Mid_Field[ rand() % nMiddle ];
  m_players[ 6 ] = g_Mid_Field[ rand() % nMiddle ];
  m_players[ 7 ] = g_Mid_Field[ rand() % nMiddle ];
  m_players[ 8 ] = g_Mid_Field[ rand() % nMiddle ];

  int nForward = (sizeof(g_Forwards ) / sizeof(g_Forwards[0] ) );
  m_players[ 9 ]  = g_Forwards[ rand() % nForward ];
  m_players[ 10 ] = g_Forwards[ rand() % nForward ];
}


/**
 * Display our team of 11 players.
 */
void CSolution::display( )
{
  int cost  = 0;
  int total = 0;

  printf( "Team | Cost | Points |    Position  | Name\n" );
  printf( "-----------------------------------------------------\n" );
  /* Total each players points + cost. */
  for( int i = 0; i < 11; i++ )
  {
    CPlayer player = m_players[i];
    printf( "%4s | %4d | %6d | %12s | %s\n",
	    player.getClub().c_str(),
	    player.getPrice(),
	    player.getPoints(),
	    player.getPosition().c_str(),
	    player.getName().c_str()
	    );


    total += player.getPoints();
    cost  += player.getPrice();
  }

  printf("==================================\n" );
  printf("Total Cost: %d  Total Points: %d\n",
	 cost, total );
}


/**
 * Breed this object with another object.
 * This is done without mutation and is a simple genome split.
 */
CSolution & CSolution::breed( CSolution partner )
{
  /*
   * Breeding with yourself is bad ;)
   */
  if ( &partner != this )
  {
      /* Pick the position to split the genome at. */
      int f = rand()%11;

      /* Splice. */
      for( int i = 0; i < f; i++ )
      {
	  CPlayer tmp          = m_players[ i ];
	  m_players[i]         = partner.m_players[i];
	  partner.m_players[i] = tmp;
      }

      /* Random mutation. */
      if ( rand()%3 == 1 )
      {
          int a = rand() % 11;
          int b = rand() % 11;

          CPlayer tmp            = m_players[ a ];
          m_players[ a ]         = partner.m_players[ b ];
          partner.m_players[ b ] = m_players[ a ];
      }
  }

  return( *this);
}
